//Relations.h
#ifndef _RELATIONS_H
#define _RELATIONS_H

#include <map>
#include <vector>
#include <algorithm>

using namespace std;


/**
        OneOne relationship template class
*/
template < class First, class Second, 
           class CompareFirst = less<First>,
           class CompareSecond = less<Second> >
class OneOne
{

typedef map <First,  Second, CompareFirst> FirstSecond;
typedef map <Second, First, CompareSecond>SecondFirst;

FirstSecond _FirstSecond;
SecondFirst _SecondFirst;

public:

/**
        Insert a pair of elements
*/
bool insert(const First& f, const Second& s){


if(_FirstSecond.end() == _FirstSecond.find(f) &&
   _SecondFirst.end() == _SecondFirst.find(s)){

  _FirstSecond[f] = s;
  _SecondFirst[s] = f;
  return true;

}

return false;

}


/**
        Remove a pair with this First if it was inserted
*/
bool removeByFirst(const First& f){

if(_FirstSecond.end() == _FirstSecond.find(f))
  return false;

Second & s = _FirstSecond[f];
_SecondFirst.erase(s);

_FirstSecond.erase(f);

return true;


}


/**
        Remove a pair with this Second if it was inserted
*/
bool removeBySecond(const Second& s){

if(_SecondFirst.end() == _SecondFirst.find(s))
  return false;

First & f = _SecondFirst[s];
_FirstSecond.erase(f);

_SecondFirst.erase(s);

return true;

}


/**
        Return the Second item for this First
        or NULL if it doesn't exist
*/
Second * getSecond(const First& f){

FirstSecond::iterator it = _FirstSecond.find(f);

return (_FirstSecond.end() == it) ? NULL : &(it->second);

}

/**
        Return the First item for this Second
        or NULL if it doesn't exist
*/
First * getFirst(const Second& s){

SecondFirst::iterator it = _SecondFirst.find(s);

return (_SecondFirst.end() == it) ? NULL : &(it->second);

}

/**
        Clear this relationship
*/
void clear(){
 _FirstSecond.clear();
 _SecondFirst.clear();

}

/**
        Define iterator for this relationship
*/
typedef FirstSecond::iterator iterator;

/**
        Return the begin iterator 
        for this relationship
*/
iterator begin() { return _FirstSecond.begin(); }

/**
        Return the end iterator 
        for this relationship
*/
iterator end() { return _FirstSecond.end(); }

};

/**
        OneMany relationship remplate class
*/

template <class One,
          class Many,
          class CompareOne = less<One>,
          class CompareMany= less<Many> >

class OneMany
{

//maps One to several Many
typedef multimap <One, Many, CompareOne> OneManyMap;
OneManyMap _OneManyMap;

//maps each Many to its One
typedef map <Many, One, CompareMany> ManyOneMap;
ManyOneMap _ManyOneMap;


public:
/**
        Insert this One Many pair to this relationship
        if it is not already there
*/
bool insert(const One& o, const Many& m){

//Does this many have a one?
if(_ManyOneMap.end() != _ManyOneMap.find(m))
    return false;


_OneManyMap.insert( pair<One, Many>(o, m) );
_ManyOneMap.insert( pair<Many, One>(m, o) );


return true;

}


/**
        Remove this Many if it was inserted
*/
bool removeByMany(const Many& m){

ManyOneMap::iterator pOne = _ManyOneMap.find(m);
if(_ManyOneMap.end() == pOne)
  return false;

_ManyOneMap.erase(m);

typedef OneManyMap::iterator MI;
pair<MI, MI> g = _OneManyMap.equal_range(pOne->second);

CompareMany pred;
for(MI p = g.first; p != g.second; ++p)
  if(!pred(p->second, m) && !pred(m, p->second)){
    _OneManyMap.erase(p);
    break;
  }

return true;

}


/**
        Remove this One if it was inserted
*/
bool removeByOne(const One& o){

typedef OneManyMap::iterator MI;
pair<MI, MI> g = _OneManyMap.equal_range(o);

//not found
if(g.first == g.second)
  return false;

for(MI p = g.first; p != g.second; ++p){
  ManyOneMap::iterator i = _ManyOneMap.find(p->second);
  if(i != _ManyOneMap.end())
     _ManyOneMap.erase(i);
  
 }

_OneManyMap.erase(o);
   
return true;

}


/**
        Return the One for this Many
        or NULL if it doesn't exist
*/

One* getOne(const Many& m){

ManyOneMap::iterator i = _ManyOneMap.find(m);

return (i == _ManyOneMap.end()) ? NULL : &(i->second);


}


/**
        Iterator class for Many 
*/
class ManysIterator{

protected:
typedef OneManyMap::iterator OMI;
OMI _mi;

typedef bidirectional_iterator_tag iterator_category;
typedef ManysIterator<One, Many> self;

public:

  ManysIterator(OMI mi = OMI()) : _mi(mi){}
  ManysIterator(const self& other) : _mi(other._mi){}

  bool operator==(const self& x) const { return _mi == x._mi; }
  bool operator!=(const self& x) const { return _mi != x._mi; }

  Many& operator*() const { return _mi->second; }


  Many* operator->() const { return &(operator*()); }


  self& operator++() {
    _mi++;
    return *this;
  }
  
  self operator++(int) {
    self tmp = *this;
    ++*this;
    return tmp;
  }
  
  self& operator--() {
    _mi--;
    return *this;
  }
  
  self operator--(int) {
    self tmp = *this;
    --*this;
    return tmp;
  }

};

typedef ManysIterator <One, Many> iterator;


class ConstManysIterator : public ManysIterator{

protected:
typedef OneManyMap::const_iterator OMI;
OMI _mi;

typedef bidirectional_iterator_tag iterator_category;
typedef ConstManysIterator<One, Many> self;

public:

  ConstManysIterator(OMI mi = OMI()) : _mi(mi){}
  ConstManysIterator(const self& other) : _mi(other._mi){}

  bool operator==(const self& x) const { return _mi == x._mi; }
  bool operator!=(const self& x) const { return _mi != x._mi; }

  const Many& operator*() const { return _mi->second; }


  const Many* operator->() const { return &(operator*()); }


  self& operator++() {
    _mi++;
    return *this;
  }
  
  self operator++(int) {
    self tmp = *this;
    ++*this;
    return tmp;
  }
  
  self& operator--() {
    _mi--;
    return *this;
  }
  
  self operator--(int) {
    self tmp = *this;
    --*this;
    return tmp;
  }

};

typedef ConstManysIterator <One, Many> const_iterator;



/**
        The begin iterator for this One
*/
iterator begin(const One& o) { return iterator(_OneManyMap.lower_bound(o)); }

/**
        The end iterator for this One
*/
iterator end(const One& o)   { return iterator(_OneManyMap.upper_bound(o)); }



/**
        The begin iterator for this One
*/
const_iterator begin(const One& o) const { return const_iterator(_OneManyMap.lower_bound(o)); }

/**
        The end iterator for this One
*/
const_iterator end(const One& o) const { return const_iterator(_OneManyMap.upper_bound(o)); }



/**
        Clear this relationship
*/
void clear(){

_OneManyMap.clear();
_ManyOneMap.clear();

}

};



/**
        triple struct
        used by ManyMany relationship
*/

template <class T1, class T2, class T3>
struct triple : public pair<T1, T2>{

triple(const T1& x, const T2& y, const T3& z) : pair<T1, T2>(x, y), 
                                                third(z){} 
T3 third;

};

template <class T1, class T2, class T3>
inline triple<T1, T2, T3> make_triple(const T1& x, const T2& y, const T3& z){
    return triple<T1, T2, T3>(x, y, z);
}


/**
        ValToType struct to determine type; used at compile time
*/
template <class T1, class T2, class T3>
struct ValToType{
typedef triple <T1, T2, T3> type;

};

/**
        Partial template instantiation of ValToType class
*/
template <class T1, class T2>
struct ValToType<T1, T2, void>{
typedef pair <T1, T2> type;

};



/**
        ManyMany relationship template class
*/
template <class Many1, class Many2,
          class CompareMany1 = less<Many1>,
          class CompareMany2 = less<Many2>,
          class Rel          = void>
class ManyMany{

typedef ValToType< Many1, Many2, Rel >::type  Relation;
typedef vector<Relation> Relations;
Relations _Relations;

typedef multimap <Many1, Relation, CompareMany1> One1Many2Map;
One1Many2Map _One1Many2Map;

typedef multimap <Many2, Relation, CompareMany2> One2Many1Map;
One2Many1Map _One2Many1Map;


public:

/**
        Insert this Many1 Many2 pair to the relationship
*/
bool insert(const Many1& m1, const Many2& m2){

Relation rel(m1, m2);
_Relations.push_back(rel);

_One1Many2Map.insert( make_pair (m1, rel) );
_One2Many1Map.insert( make_pair (m2, rel) );

return true;
       
}

/**
        Insert this Many1 Many2 Rel triple to the relationship
*/
bool insert(const Many1& m1, const Many2& m2, const Rel& r){

Relation rel(m1, m2, r);
_Relations.push_back(rel);

_One1Many2Map.insert( make_pair (m1, rel) );
_One2Many1Map.insert( make_pair (m2, rel) );

return true;
       
}


/**
        Remove this Many1
*/
bool removeByMany1(const Many1& m1){

CompareMany1 pred;
for(Relations::iterator it = _Relations.begin(), 
                       end = _Relations.end(); it != end; ++it){

  if(!pred(m1, it->first) && !pred(it->first, m1)){
    _One1Many2Map.erase(it->first);    

    _Relations.erase(it);

    typedef One2Many1Map::iterator MI;
    pair<MI, MI> g = _One2Many1Map.equal_range(it->second);

    for(MI p = g.first; p != g.second; ++p)
      if(!pred(p->second->first, m1) && !pred(m1, p->second->first))
        _One2Many1Map.erase(p);
    
  }
      
}

return true;

}


/**
        Remove this Many2
*/
bool removeByMany2(const Many2& m2){

CompareMany2 pred;
for(Relations::iterator it = _Relations.begin(), 
                       end = _Relations.end(); it != end; ++it){

  if(!pred(m2, it->second) && !pred(it->second, m2)){
    _One2Many1Map.erase(it->second);    

    _Relations.erase(it);

    typedef One1Many2Map::iterator MI;
    pair<MI, MI> g = _One1Many2Map.equal_range(it->first);

    for(MI p = g.first; p != g.second; ++p)
      if(!pred(p->second->second, m2) && !pred(m2, p->second->second))
        _One1Many2Map.erase(p);
    
  }
      
}

return true;

}

/**
        Functor struct used to find relations
*/
struct eqRel{

const Relation& _r;
eqRel(const Relation& r) : _r(r){}

bool operator() (const Relation& r1){

CompareMany1 pred1;
CompareMany2 pred2;

  return !pred1(r1.first,  _r.first)  && !pred1(_r.first,  r1.first) &&
         !pred2(r1.second, _r.second) && !pred2(_r.second, r1.second);

}

};


/**
        Remove this Many1 Many2 pair
*/
bool remove(const Many1& m1, const Many2& m2){
  
  Relation rem(m1, m2);
  Relations::iterator it = find_if(_Relations.begin(), _Relations.end(), eqRel(rem));
  if(_Relations.end() == it) return false;

  _Relations.erase(it);

  typedef One1Many2Map::iterator MI1;
  pair<MI1, MI1> g1 = _One1Many2Map.equal_range(m1);

  CompareMany2 pred2;
  for(MI1 p = g1.first; p != g1.second; ++p)
    if(!pred2(p->second.second, m2) && !pred2(m2, p->second.second)){
       _One1Many2Map.erase(p);
       break;
    }
 

  typedef One2Many1Map::iterator MI2;
  pair<MI2, MI2> g2 = _One2Many1Map.equal_range(m2);

  CompareMany1 pred1;
  for(MI2 p = g2.first; p != g2.second; ++p)
    if(!pred1(p->second.first, m1) && !pred1(m1, p->second.first)){
      _One2Many1Map.erase(p);
      break;
    }
  
  return true;

}

/**
        Return the Link Attribute for this pair of Many1 Many2
        or NULL if not found
*/
const Rel* getAttribute(const Many1& m1, const Many2& m2){

  Relation rem(m1, m2, Rel());
  Relations::iterator it = find_if(_Relations.begin(), 
                                       _Relations.end(), eqRel(rem));
  
  return (_Relations.end() == it) ? NULL :
                                                                    &(it->third); 

}

class Manys1Iterator{

protected:
typedef One2Many1Map::iterator OMI;
OMI _mi;

typedef bidirectional_iterator_tag iterator_category;
typedef Manys1Iterator<Many1, Many2> self;

public:

  Manys1Iterator(OMI mi = OMI()) : _mi(mi){}
  Manys1Iterator(const self & other) : _mi(other._mi){}

  bool operator==(const self& x) const { return _mi == x._mi; }
  bool operator!=(const self& x) const { return _mi != x._mi; }

  Many1 & operator*() const { return _mi->second.first; }


  Many1 * operator->() const { return &(operator*()); }


  self& operator++() {
    _mi++;
    return *this;
  }
  
  self operator++(int) {
    self tmp = *this;
    ++*this;
    return tmp;
  }
  
  self& operator--() {
    _mi--;
    return *this;
  }
  
  self operator--(int) {
    self tmp = *this;
    --*this;
    return tmp;
  }

};

class Manys2Iterator{

protected:
typedef One1Many2Map::iterator OMI;
OMI _mi;

typedef bidirectional_iterator_tag iterator_category;
typedef Manys2Iterator<Many1, Many2> self;

public:

  Manys2Iterator(OMI mi = OMI()) : _mi(mi){}
  Manys2Iterator(const self & other) : _mi(other._mi){}

  bool operator==(const self& x) const { return _mi == x._mi; }
  bool operator!=(const self& x) const { return _mi != x._mi; }

  Many2 & operator*() const { return _mi->second.second; }


  Many2 * operator->() const { return &(operator*()); }


  self& operator++() {
    _mi++;
    return *this;
  }
  
  self operator++(int) {
    self tmp = *this;
    ++*this;
    return tmp;
  }
  
  self& operator--() {
    _mi--;
    return *this;
  }
  
  self operator--(int) {
    self tmp = *this;
    --*this;
    return tmp;
  }

};

/**
        Iterator type to iterate through Many1's 
*/
typedef Manys1Iterator <Many1, Many2> iterator1;

/**
        Iterator type to iterate through Many2's 
*/
typedef Manys2Iterator <Many1, Many2> iterator2;

/**
        The begin iterator for this Many2
*/
iterator1 begin(const Many2 & m2){ return iterator1(_One2Many1Map.lower_bound(m2)); }

/**
        The end iterator for this Many2
*/
iterator1 end(const Many2 & m2)  { return iterator1(_One2Many1Map.upper_bound(m2)); }

/**
        The begin iterator for this Many1
*/
iterator2 begin(const Many1 & m1){ return iterator2(_One1Many2Map.lower_bound(m1)); }

/**
        The end iterator for this Many1
*/
iterator2 end(const Many1 & m1)  { return iterator2(_One1Many2Map.upper_bound(m1)); }


/**
        Clear this relationship
*/
void clear(){

_One1Many2Map.clear();
_One2Many1Map.clear();
_Relations.clear();

}

};


#endif


